# PullDown -- A button with a pull down menu
# Bert Bos <bert@let.rug.nl>
# $Ident$

@class XfwfPullDown (XfwfButton)  @file=PullDown

@ A PullDown button looks like a normal button, but when it is clicked
or pressed, a menu drops down. The menu can be specified in two ways:
as a string or as a widget. If a string is given, it must be in the
correct syntax for the EzMenu(3) widget and a widget will be created
automatically. If a widget is given instead, it must be a widget that
is able to pop up: in other words: a Shell widget or a subclass of
Shell.

The most likely application of this widget is as a menu in a menu bar.
A menu bar can be created, e.g., with a RowCol widget and a number of
PullDown buttons.

By default, the |activate| that is inherited from Button is
still called, but as soon as the button is first pressed, not only
after a click. Most applications will not need this callback.



@PUBLIC

@ Two resources determine the menu. The |popup| resource can be
pointed at the widget that should pop up when the button is pressed,
but when the |popup| resource is NULL, the menu will be created
>from the |menu| resource. If the |popup| resource is set, it must
be set to a widget that is (a subclass of) a Shell widget.

	@var Widget popup = NULL

@ When |popup| is |NULL|, the widget will be created from the
textual description in |menu|. For the syntax of this string, see the
EzMenu(3).

	@var String menu = NULL

@ The pull down button installs an event handler that catches all
keyboard events for the whole application and checks for a special key
combination. If that key is detected, the menu will be popped up.
Thus, the menu can be used even if the keyboard focus is somewhere
else. The |hotkey| resource is usually a string such as |"Alt<Key>a"|,
or |"Ctrl<Key>f"|.

When the |hotkey| resource is used, any |accelerators| will be ignored.

	@var String hotkey = NULL

@ The cursor shape that is used in the menu can be set with the
|menuCursor| resource. By default it is an arrow pointing to the upper
right.

	@var Cursor menuCursor = <String> "arrow"

@ Since the most likely place for this type of button will be in a
menubar, the default frame width is reset to zero.

	@var frameWidth = 0

@ The |shrinkToFit| resource is set to a new default of |True|.

	@var shrinkToFit = True



@TRANSLATIONS

@trans <Btn1Down>: open_menu()
@trans Button1<Enter>: open_menu()
@trans <Key>Return: open_menu()



@ACTIONS

@ |note| is for debugging only.

@proc note
{
    static String Xeventname[] = {
	"", "", "KeyPress", "KeyRelease", "ButtonPress", "ButtonRelease",
	"MotionNotify", "EnterNotify", "LeaveNotify", "FocusIn",
	"FocusOut", "KeymapNotify", "Expose", "GraphicsExpose",
	"NoExpose", "VisisbilityNotify", "CreateNotify", "DestroyNotify",
	"UnmapNotify", "MapNotify", "MapRequest", "ReparentNotify",
	"ConfigureNotify", "ConfigureRequest", "GravityNotify",
	"ResizeRequest", "CirculateNotify", "CirculateRequest",
	"PropertyNotify", "SelectionClear", "SelectionRequest",
	"SelectionNotify", "ColormapNotify", "ClientMessage",
	"MappingNotify", };
    (void) fprintf(stderr, "%s received %s\n",
		   XtName($), Xeventname[event->type]);
}


@ The |open_menu| action computes the coordinates of the menu, such
that it will appear directly below the button. It pops up the menu and
calls the |activate| routines.

@proc open_menu
{
    Position x, y;

    if ($popup == NULL) return;
    XtTranslateCoords($, 0, $height, &x, &y);
    XtMoveWidget($popup, x, y);
    XtPopupSpringLoaded($popup);
    XtGrabPointer($popup, True, EnterWindowMask | LeaveWindowMask |
		  ButtonPressMask | ButtonReleaseMask, GrabModeAsync,
		  GrabModeAsync, None, $cursor, CurrentTime);
    XtCallCallbackList($, $activate, event);
}



@METHODS

@ Both the |initialize| and |set_values| methods need to create a new
menu widget, if the |menu| resource is set and the |popup| resource
is not. A utility function is used to do the actual work. If both
|popup| and |menu| are set, |popup| is used and |menu| is
ignored.

@proc initialize
{
    if ($popup == NULL && $menu != NULL) create_menu($);
    if ($popup != NULL && $hotkey != NULL) create_accelerator($);
}


@ |set_values| is similar. Changes in these resources never cause a
redraw, so the method always returns |False|.

@proc set_values
{
    if ($old$popup != $popup)
	XtDestroyWidget($old$popup);
    else if ($old$menu != $menu)
	create_menu($);
    if ($old$menuCursor != $menuCursor)
	if ($popup) XtVaSetValues($popup, XtNcursor, $menuCursor, NULL);
    if ($old$hotkey != $hotkey)
	create_accelerator($);
    return False;
}



@UTILITIES

@ The |create_menu| routine is used to create an instance of an EzMenu
(pronounced `e-ZEE-menu', an easy-to-use version of the Athena simple
menu).

@proc create_menu($)
{
    if ($popup) XtDestroyWidget($popup);
    $popup = XtVaCreatePopupShell("_menu", xfwfEzMenuWidgetClass, $,
				  XtNmenu, $menu, XtNcursor, $menuCursor,
				  NULL);
}


@ |create_accelerator| takes the |hotkey| and expands it into a
correct |accelerators| resource.

@proc create_accelerator($)
{
    char s[100];
    XtAccelerators a;

    (void) strcpy(s, $hotkey);
    (void) strcat(s, ": open_menu()");
    a = XtParseAcceleratorTable(s);
    XtVaSetValues($, XtNaccelerators, a, NULL);
}



@IMPORTS

@incl <Xfwf/EzMenu.h>

@incl <stdio.h>
