@class XfwfToggle(XfwfButton)  @file=Toggle

@ The |XfwfToggle| is a button that switches states with every activation
(which is by default with every mouse click). The states are named `on' and
`off'. The states can be indicated with a $\surd$ before the label Two
callbacks report the changed state to the application: |onCallback| is called
when the button switches to `on', |offCallback| is called when the button
switches back to `off'.


@public

@ The |onCallback| is called by the |toggle| action or by the |switch_on|
action, if the previous state was `off'. The |call_data| argument will contain
the |XEvent| pointer that trigerred the action function.

@var <Callback> XtCallbackList onCallback = NULL

@ The |offCallback| is called from the |toggle| or |switch_off| action, if the
previous state was `on'. The |call_data| argument will be a pointer to the
|XEvent| that caused the action.

@var <Callback> XtCallbackList offCallback = NULL

@ The |tickmark| resource controls whether the label in the button will be
prefixed with a tick mark $\surd$ in the `on' state. If |tickmark| is |False|,
the application should find some other means of indicating the state, e.g, by
changing the label.

@var Boolean tickmark = True

@ The variable |on| records the state of the widget: |True| means `on' and
|False| means `off'.

@var Boolean on = False


@translations

@ The |toggle| action toggles the widget between `on' and `off'. By
default it is bound to a click of the left mouse button as well as to
the Return key.

@trans <Btn1Down>,<Btn1Up>: toggle()
@trans <Key>Return: toggle()


@actions

@ The |toggle| action switches the state. Depending on the resources |tickmark|
and |invert| it might change the label or the colors or both. The |onCallback|
or |offCallback| functions are called, with the event as |call_data| argument.

@proc toggle
{
    XtVaSetValues($, "on", !$on, NULL);
    XtCallCallbackList($, $on ? $onCallback : $offCallback, event);
}

@ The |switch_on| action switches the button to `on' if it is `off', otherwise
it does nothing. By default it isn't bound to any event. If the widget is
changed, the |onCallback| is called with the event as |call_data|.

@proc switch_on
{
    if (! $on) {
        XtVaSetValues($, "on", True, NULL);
        XtCallCallbackList($, $onCallback, event);
    }
}

@ The |switch_off| action switches the widget to `off' if the state is `on',
otherwise it does nothing. When the widget changes states, the |offCallback|
is called, with a pointer to the |XEvent| structure as |call_data| argument.

@proc switch_off
{
    if ($on) {
        XtVaSetValues($, "on", False, NULL);
        XtCallCallbackList($, $offCallback, event);
    }
}


@private

@ The |tickgc| GC holds the picture of a tick mark as a stipple.

@var GC tickgc


@methods

@ The tick mark is a small bitmap that is created in a utility function
and put into a GC called |tickgc|.

@proc initialize
{
    $tickgc = NULL;
    create_tickgc($);
}

@ When the |on| resource changes, the |set_values| method inspects the
|tickmark| resource to see if the widget needs to be redrawn (to add or remove
the tick mark).

@proc set_values
{
    if ($foreground != $old$foreground)
        create_tickgc($);
    if ($on != $old$on && $tickmark)
        return True;
    else
        return False;
}

@ The |expose| method uses the |expose| method of the superclass to draw the
button and then possibly adds a tick mark.

@proc expose
{
    Position x, y;
    Dimension w, h;

    if (! XtIsRealized($)) return;
    #expose($, event, region);
    if ($tickmark && $on) {
        $compute_inside($, &x, &y, &w, &h);
        y = y + (h - tick_height)/2;
	XSetTSOrigin(XtDisplay($), $tickgc, x, y);
        XFillRectangle(XtDisplay($), XtWindow($), $tickgc, x, y,
            tick_width, tick_height);
    }
}


@utilities

@ The |create_tickgc| function creates a GC with the tick mark bitmap as a
stipple and the |foreground_pixel| as the foreground.

@proc create_tickgc($)
{
    XtGCMask mask;
    XGCValues values;

    if ($tickgc != NULL) XtReleaseGC($, $tickgc);
    mask = GCForeground | GCStipple | GCFillStyle;
    values.foreground = $foreground;
    values.stipple = XCreateBitmapFromData(XtDisplay($),
        RootWindowOfScreen(XtScreen($)), tick_bits, tick_width, tick_height);
    if (! values.stipple) XtError("Couldn't create bitmap for tick mark");
    values.fill_style = FillStippled;
    $tickgc = XtGetGC($, mask, &values);
}


@imports

@incl "tick.bm"

